@@ -22,9 +22,10 @@
 //#####################################################################
 //#####################################################################
 
-	static int liqbookreader_paint_selfrotate(liqcell *self, liqcellpainteventargs *args, struct doc *doc)
-	{
-		
+	static int liqbookreader_paint_selfrotate(liqcell *self, liqcellpainteventargs *args,void *context)
+{
+
+		struct doc *doc = (struct doc *)liqcell_gettag(self);
 		
 		// the body contains the offset we need to render at
 		// uber render, GOGOGO!
@@ -97,14 +98,33 @@
 
 
 
-	static int liqbookreader_paint(liqcell *self, liqcellpainteventargs *args, struct doc *doc)
-	{
+	static int liqbookreader_paint(liqcell *self, liqcellpainteventargs *args, void *context)
+{
+
+		struct doc *doc = (struct doc *)liqcell_gettag(self);
 		
 		// the body contains the offset we need to render at
 		// uber render, GOGOGO!
 		liqcell *body = liqcell_child_lookup(self, "body");
 		
 		
+		// find out if orientation flipped.
+		if( canvas.pixelwidth != liqcell_getw(body) )
+		{
+			// something changed, a glitch in the matrix.
+			char *bookfilename = liqcell_propgets(self,"book_filename",NULL);
+			unsigned int sfl = body->h;
+			unsigned int sfu = -body->y;
+			liqapp_log("book dynamic relayout, sfu=%d, sfl=%d",sfu,sfl);
+			liqbookprogress_finishedreading_storeposition(bookfilename,sfl,sfu);
+			liqcell_handlerrun(self,"layout",NULL);
+			liqcell_setdirty(self,1);
+			return 0;
+			
+		}
+		
+		
+		
 		//if( -liqcell_gety(body) + liqcell_geth(body)   )
 		
 		//liqcell_setpos(body, liqcell_getx(body)-10,liqcell_getx(body) );
@@ -122,6 +142,7 @@
 		//int		vgraph_setfont(        			vgraph *self, liqfont *font);			//  char *fontname, int fontsize, int fontattributes
 		//int		vgraph_setpencolor(    			vgraph *self, vcolor pencolor );
 		//int		vgraph_drawtext(       			vgraph *self, int x, int y, char *text        );
+		liqfont_setview(doc->renderfont, 1,1);
 		vgraph_setfont(args->graph,      doc->renderfont);
 		vgraph_setpencolor( args->graph, vcolor_YUV(255,128,128) );
 		struct docline *dline = doc->linefirst;
@@ -133,7 +154,7 @@
 		while(dline)
 		{
 			h=dline->formath;
-			//liqapp_log("huh %d %d,  vt=%d vb=%d",y,h,vt,vb);
+			liqapp_log("huh %d %d,  vt=%d vb=%d",y,h,vt,vb);
 			if(y+h > vt)
 			{
 				//liqapp_log("tmp %d",tmp);
@@ -180,19 +201,147 @@ int liqbookreader_getbookfontsize()
 }
 
 
+int liqbookreader_saveaspng(liqcell *self)
+{
+	struct doc *doc = (struct doc *)liqcell_gettag(self);
+	liqcell *body = liqcell_child_lookup(self, "body");
+	
+	char *book_filename = liqcell_propgets(self,"book_filename",NULL);
+	
+	int font_height = liqfont_textheight(doc->renderfont);
 
+	int scr_height = self->h;
+	int book_linecount=doc->linecount;
+	int book_totalheight=body->h;
+	
+	int page_topstart=-body->y;
+	int page_lineviewcount=self->h/font_height;
+	int page_topmax=book_totalheight-canvas.pixelheight;
+	
+	
+	int page_width = canvas.pixelwidth;
+	int page_height = book_totalheight;	
+	
+    liqcell_hold(self);
+
+        liqapp_log("liqbookreader_saveaspng creating image");
+        liqimage *img = liqimage_newatsize(page_width,page_height,0);
+        if(!img)
+        {
+            liqapp_log("liqbookreader_saveaspng could not alloc image");
+            liqcell_release(self);
+            return -1;
+        }
+        
+        liqapp_log("liqbookreader_saveaspng creating cliprect");
+        liqcliprect *cr = liqcliprect_newfromimage(img);
+        if(!cr)
+        {
+            liqapp_log("liqbookreader_saveaspng could not alloc cr");
+            
+            liqimage_release(img);
+            liqcell_release(self);
+            return -1;
+        }
+
+
+		char indexfilename[FILENAME_MAX]={0};
+		FILE *indexfile;
+		snprintf(indexfilename,sizeof(indexfilename),"%s.LiqBook.html",book_filename);
+		indexfile=fopen(indexfilename,"w");
+		if(!indexfile)
+		{
+            liqapp_log("liqbookreader_saveaspng could not open indexfile");
+            
+            liqimage_release(img);
+            liqcell_release(self);
+            return -1;
+		}
+		
+		liqapp_log("liqbookreader_saveaspng setting up index");
+
+		fprintf(indexfile,"<HTML>\n");
+		fprintf(indexfile," <HEAD>\n");
+		fprintf(indexfile," <TITLE>LiqBook '%s' by Gary Birkett.</TITLE>\n", liqapp_filename_walkoverpath(book_filename) );
+		fprintf(indexfile," </HEAD>\n");
+		fprintf(indexfile," <BODY bgcolor=\"#000000\" text=\"#ffffff\">\n");
+
+		
+		// Now, cycle the pages
+		
+		self->h = body->h;
+		body->y = 0;
+		
+		int lineupto = 0;
+		int pagenumber=0;
+
+		
+		char pagefilename[FILENAME_MAX]={0};
+		//for(lineupto = 0; lineupto < book_linecount; lineupto+=page_lineviewcount )
+		{
+			body->y = -(font_height * page_lineviewcount) * pagenumber;
+			
+			
+			snprintf(pagefilename,sizeof(pagefilename),"%s.page.%03d.png",book_filename,pagenumber);
+    
+			fprintf(indexfile," <img src='%s' width='%d' height='%d'></img><BR/>\n",liqapp_filename_walkoverpath(pagefilename), page_width,page_height);
+			liqapp_log("liqbookreader_saveaspng clearing before cell %s",pagefilename);
+			liqcliprect_drawclear(cr,0,128,128);
+			
+			liqapp_log("liqbookreader_saveaspng painting cell %s",pagefilename);
+			liqcell_easypaint(self,cr,0,0,page_width,page_height);
+			liqapp_log("liqbookreader_saveaspng saving image as '%s'",pagefilename);
+			liqimage_pagesavepng(img,pagefilename);
+			
+			pagenumber++;
+		}
+		
+		
+
+
+
+		fprintf(indexfile," <A href='http://liqbase.net'>\n");
+		fprintf(indexfile,"  <P>Created by Gary Birkett,</P>\n");
+		fprintf(indexfile,"  <P>Smile Architect.</P>\n");
+		fprintf(indexfile," </A>\n");
+		fprintf(indexfile," </BODY>\n");
+		fprintf(indexfile,"<HTML>\n");
+		fclose(indexfile);
+	
+
+
+    
+        
+        liqapp_log("liqbookreader_saveaspng releasing cr");
+        liqcliprect_release(cr);
+        
+        liqapp_log("liqbookreader_saveaspng releasing image");
+        liqimage_release(img);
+	
+	liqapp_log("liqbookreader_saveaspng done");
+    
+	
+	body->y = page_topstart;
+	self->h = scr_height;
+	
+    liqcell_release(self);
+}
 
 
 /**	
- * liqbookreader_cover keypress - the user pressed a key
+ * liqbookreader keypress - the user pressed a key
  */	
-static int liqbookreader_cover_keypress(liqcell *self, liqcellkeyeventargs *args,struct doc *doc)
+static int liqbookreader_keypress(liqcell *self, liqcellkeyeventargs *args,void *context)
 {
+
+		struct doc *doc = (struct doc *)liqcell_gettag(self);
 	liqcell *body = liqcell_child_lookup(self, "body");
 	
 	int font_height = liqfont_textheight(doc->renderfont);
 	
+	int book_linecount=doc->linecount;
 	int book_totalheight=body->h;
+	
 	int page_topstart=-body->y;
 	int page_lineviewcount=self->h/font_height;
 	int page_topmax=book_totalheight-canvas.pixelheight;
@@ -202,7 +351,21 @@ static int liqbookreader_cover_keypress(liqcell *self, liqcellkeyeventargs *args
 	{
 		liqapp_pref_setvalue_printf( "book_fontsize", "%i", liqbookreader_getbookfontsize()-2 );
 		liqapp_prefs_save();
-		liqcell_handlerrun(self,"refresh",NULL);
+		liqcell_handlerrun(self,"layout",NULL);
+		// smaller font
+	}
+	else
+	if(args->keycode==115)		// s
+	{
+		// saving book as png.
+		// allocate "page"
+		// render to it.
+		
+		liqbookreader_saveaspng(self);
+		
+		//liqapp_pref_setvalue_printf( "book_fontsize", "%i", liqbookreader_getbookfontsize()-2 );
+		//liqapp_prefs_save();
+		//liqcell_handlerrun(self,"layout",NULL);
 		// smaller font
 	}
 	else
@@ -211,7 +374,7 @@ static int liqbookreader_cover_keypress(liqcell *self, liqcellkeyeventargs *args
 		
 		liqapp_pref_setvalue_printf( "book_fontsize", "%i", liqbookreader_getbookfontsize()+2 );
 		liqapp_prefs_save();
-		liqcell_handlerrun(self,"refresh",NULL);
+		liqcell_handlerrun(self,"layout",NULL);
 		// bigger font
 	}
 	else
@@ -254,10 +417,12 @@ static int liqbookreader_cover_keypress(liqcell *self, liqcellkeyeventargs *args
 	return 0;
 }
 /**	
- * liqbookreader_cover keyrelease - the user released a key
+ * liqbookreader keyrelease - the user released a key
  */	
-static int liqbookreader_cover_keyrelease(liqcell *self, liqcellkeyeventargs *args,struct doc *doc)
+static int liqbookreader_keyrelease(liqcell *self, liqcellkeyeventargs *args,void *context)
 {
+
+		struct doc *doc = (struct doc *)liqcell_gettag(self);
 	return 0;
 }
 
@@ -265,11 +430,18 @@ static int liqbookreader_cover_keyrelease(liqcell *self, liqcellkeyeventargs *ar
 
 
 /**	
- * liqbookreader refresh - rebuild book details
+ * liqbookreader layout - rebuild book details
  */	
-static int liqbookreader_refresh(liqcell *self, liqcelleventargs *args,struct doc *doc)
+static int liqbookreader_layout(liqcell *self, liqcelleventargs *args,void *context)
 {
 
+		struct doc *doc = (struct doc *)liqcell_gettag(self);
+
+		liqcell_setrect(self, 0,0, canvas.pixelwidth,canvas.pixelheight);
+
+		
+		liqcell *body = liqcell_child_lookup(self, "body");
+		
 	
 		int err=0;
 		
@@ -281,18 +453,18 @@ static int liqbookreader_refresh(liqcell *self, liqcelleventargs *args,struct do
 		
 		//=================================================== now, open the font
 		
-		liqfont *font = liqfont_cache_getttf("/usr/share/fonts/nokia/nosnb.ttf", (edfontsize), 0);
+		liqfont *font = liqfont_newfromfilettf("/usr/share/fonts/nokia/nosnb.ttf", (edfontsize), 0);
 
 
 		if(!font)
 		{
-			{ return liqapp_warnandcontinue(-1,"liqbookreader_run couldnt open font"); }						
+			{ return liqapp_warnandcontinue(-1,"liqbookreader_layout couldnt open font"); }						
 		}
 		
 		//liqfont_setscale(font,1,1);
 
 
-
+		liqfont_setview(font, 1,1);
 
 		//=================================================== initialize the bookprogress
 
@@ -302,12 +474,12 @@ static int liqbookreader_refresh(liqcell *self, liqcelleventargs *args,struct do
 
 		if(doc->renderfont)
 		{
-			liqfont_cache_release(doc->renderfont);
+			liqfont_release(doc->renderfont);
 			doc->renderfont=NULL;
 		}
 		
 		doc->renderfont=font;
-		doc->rendermarginright=480;
+		doc->rendermarginright=canvas.pixelwidth;
 		//doc.rendermarginbottom=99999999;
 	
 		err=doc_initfromfilename(doc,bookfilename);
@@ -315,14 +487,13 @@ static int liqbookreader_refresh(liqcell *self, liqcelleventargs *args,struct do
 		
 		if(err)
 		{
-			{ return liqapp_warnandcontinue(-1,"liqbookreader_run couldnt open doc"); }						
+			{ return liqapp_warnandcontinue(-1,"liqbookreader_layout couldnt open doc"); }						
 		}
-		
-		liqcell *body = liqcell_child_lookup(self, "body");
-		
-		
-		liqcell_setrect(body, 0,0, 480,doc->rendertoth);
 
+
+
+		liqcell_setrect(self, 0,0, canvas.pixelwidth,canvas.pixelheight);
+		liqcell_setrect(body, 0,0, canvas.pixelwidth,doc->rendertoth);
 		
 		unsigned int sfl=0;
 		unsigned int sfu=0;
@@ -330,11 +501,11 @@ static int liqbookreader_refresh(liqcell *self, liqcelleventargs *args,struct do
 		{
 			//
 			if(sfu>sfl-(self->h))sfu=sfl-self->h;
-			liqapp_log("book opening, sfu=%d, sfl=%d",sfu,sfl);
+			liqapp_log("book layout, sfu=%d, sfl=%d",sfu,sfl);
 			if(body->h>0 && sfl>0)
 				liqcell_setpos(body,0, -( sfu * (body->h/128) / (sfl/128)));
 				
-			liqapp_log("book opening, body->y=%d, body->h=%d",body->y,body->h);
+			liqapp_log("book layout, body->y=%d, body->h=%d",body->y,body->h);
 				
 		}
 		
@@ -353,7 +524,7 @@ static int liqbookreader_refresh(liqcell *self, liqcelleventargs *args,struct do
 
 int liqbookreader_run(char *bookfilename)
 {
-	liqcell *self = liqcell_quickcreatewidget("liqbookreader","form", 480,800);
+	liqcell *self = liqcell_quickcreatewidget("liqbookreader","form", canvas.pixelwidth,canvas.pixelheight);
 
 	if(self)
 	{
@@ -365,125 +536,40 @@ int liqbookreader_run(char *bookfilename)
 		int err=0;
 		
 		//char *bookfilename = liqcell_propgets(self,"bookfilename",NULL);
+	
+
+
+		liqcell *body = liqcell_quickcreatevis("body",NULL,0 ,0, canvas.pixelwidth,canvas.pixelheight  );
+		liqcell_handleradd(body,    "mouse",   liqcell_easyhandler_kinetic_mouse );
+		liqcell_child_insert( self, body );
 		
-		//================================================== get and process the fontsize
-		int edfontsize=liqbookreader_getbookfontsize();
-		
-		liqapp_log("liqbookreader_run act GB fontsize = %d",edfontsize);
-		
-		//=================================================== now, open the font
-		
-		liqfont *font = liqfont_cache_getttf("/usr/share/fonts/nokia/nosnb.ttf", (edfontsize), 0);
 
 
-		if(!font)
-		{
-			{ return liqapp_warnandcontinue(-1,"liqbookreader_run couldnt open font"); }						
-		}
-		
-		//liqfont_setscale(font,1,1);
-
-
-
-
-		//=================================================== initialize the bookprogress
-
-		liqbookprogress_shutdown();
-		liqbookprogress_startup();
-
-
-
-
-
-
-		//=================================================== open the book now 
-
-		// grab the file and process it :)
+		// create base doc structure
 		struct doc doc;
 		doc.renderfont=NULL;
 		
 		
-		doc.renderfont=font;
-		doc.rendermarginright=480;
-		//doc.rendermarginbottom=99999999;
-	
-
-
-		err=doc_initfromfilename(&doc,bookfilename);
-		
-		
-		if(err)
-		{
-			{ return liqapp_warnandcontinue(-1,"liqbookreader_run couldnt open doc"); }						
-		}
-			
-		
-	//	// change to rotated now we have weighed it
-	//	liqfont_release(font);
-	//	font = liqfont_cache_getttf("/usr/share/fonts/nokia/nosnb.ttf", (edfontsize), 90);
-	//	doc.renderfont=font;
-		
-		
-		liqcell *body = liqcell_quickcreatevis("body",NULL,0 ,0, 480,doc.rendertoth  );
-		liqcell_handleradd(body,    "mouse",   liqcell_easyhandler_kinetic_mouse );
-		liqcell_child_insert( self, body );
+		// link them badly, this should be a tag*
+		liqcell_settag(self,(int)&doc);
 		
 		
 		//liqcell_handleradd_withcontext(body,    "paint", widget_paint,self);
 		
-		
-		unsigned int sfl=0;
-		unsigned int sfu=0;
-		if(liqbookprogress_startreading_lookupposition(bookfilename,&sfl,&sfu))
-		{
-			//
-			if(sfu>sfl-(self->h))sfu=sfl-self->h;
-			liqapp_log("book opening, sfu=%d, sfl=%d",sfu,sfl);
-			if(body->h>0 && sfl>0)
-				liqcell_setpos(body,0, -( sfu * (body->h/128) / (sfl/128)));
-				
-			liqapp_log("book opening, body->y=%d, body->h=%d",body->y,body->h);
-				
-		}
-		
-		
-		// now, create the booklines or generate on the fly...
-	
-		//########################################## simple style..
-		/*
-		struct docline *dline = doc.linefirst;
-		int y=0;
-		int usedcount=0;
-		while(dline)
-		{
-
-			// no paragraph grouping, very slow for >10k lines (normal book size)
-			liqcell *scrline = liqcell_quickcreatevis(dline->linedata,NULL,0 ,y,   800,dline->formath);
-			liqcell_setfont(scrline,liqfont_hold(font));
-			liqcell_child_insert( body, scrline );
-						
-			
-			y+=dline->formath;
-			dline=dline->linknext;
-			if(y>4000) break;
-			usedcount++;
-			
-		}
-		liqapp_log("liqbookreader added %d out of %d lines",doc.linecount,usedcount);
-		*/
-		
 		//########################################## custom renderer ..
 		
-		liqcell_handleradd_withcontext(self, "refresh",    (void*)liqbookreader_refresh,          &doc);
-		liqcell_handleradd_withcontext(self, "paint",      (void*)liqbookreader_paint,            &doc);
-		liqcell_handleradd_withcontext(self, "keypress",   (void*)liqbookreader_cover_keypress,   &doc);
-		liqcell_handleradd_withcontext(self, "keyrelease", (void*)liqbookreader_cover_keyrelease, &doc);
+		liqcell_handleradd_withcontext(self, "layout",     (void*)liqbookreader_layout,     NULL);
+		liqcell_handleradd_withcontext(self, "paint",      (void*)liqbookreader_paint,      NULL);
+		liqcell_handleradd_withcontext(self, "keypress",   (void*)liqbookreader_keypress,   NULL);
+		liqcell_handleradd_withcontext(self, "keyrelease", (void*)liqbookreader_keyrelease, NULL);
+		
+		liqcell_handlerrun(self,"layout",NULL);
 		
 		liqcell_easyrun(self);
 		
 		
-		sfl = body->h;
-		sfu = -body->y;
+		unsigned int sfl = body->h;
+		unsigned int sfu = -body->y;
 		liqapp_log("book closing, sfu=%d, sfl=%d",sfu,sfl);
 		liqbookprogress_finishedreading_storeposition(bookfilename,sfl,sfu);
 	}
